import { __spreadArray, __read } from 'tslib';
import sync, { cancelSync, flushSync } from 'framesync';
import { SubscriptionManager } from '../../utils/subscription-manager.js';
import { copyBoxInto } from '../geometry/copy.js';
import { applyBoxDelta, applyTreeDeltas } from '../geometry/delta-apply.js';
import { calcBoxDelta, calcLength } from '../geometry/delta-calc.js';
import { removeBoxTransforms } from '../geometry/delta-remove.js';
import { createDelta, createBox } from '../geometry/models.js';
import { translateAxis, transformBox } from '../geometry/operations.js';
import { isDeltaZero } from '../geometry/utils.js';
import { NodeStack } from '../shared/stack.js';
import { scaleCorrectors } from '../styles/scale-correction.js';
import { buildProjectionTransform } from '../styles/transform.js';
import { eachAxis } from '../utils/each-axis.js';
import { hasTransform, hasScale } from '../utils/has-transform.js';

function createProjectionNode(_a) {
    var attachResizeListener = _a.attachResizeListener, defaultParent = _a.defaultParent, measureScroll = _a.measureScroll, measureViewportBox = _a.measureViewportBox, resetTransform = _a.resetTransform;
    return /** @class */ (function () {
        function ProjectionNode(id, latestValues, parent) {
            var _this = this;
            if (parent === void 0) { parent = defaultParent === null || defaultParent === void 0 ? void 0 : defaultParent(); }
            this.children = new Set();
            this.options = {};
            this.isLayoutDirty = false;
            this.isUpdating = false;
            this.shouldResetTransform = false;
            this.treeScale = { x: 1, y: 1 }; // TODO Lazy-initialise
            // Note: Currently only running on root node
            this.potentialNodes = new Map();
            this.updateProjection = function () {
                updateProjectionTree(_this);
            };
            this.isVisible = true;
            /**
             * Shared layout
             */
            // TODO Only running on root node
            this.sharedNodes = new Map();
            this.id = id;
            this.latestValues = latestValues;
            this.root = parent ? parent.root || parent : this;
            this.path = parent ? __spreadArray(__spreadArray([], __read(parent.path)), [parent]) : [];
            this.parent = parent;
            this.root.registerPotentialNode(id, this);
        }
        ProjectionNode.prototype.destructor = function () {
            var _a;
            (_a = this.parent) === null || _a === void 0 ? void 0 : _a.children.delete(this);
            cancelSync.preRender(this.updateProjection);
        };
        ProjectionNode.prototype.registerPotentialNode = function (id, node) {
            this.potentialNodes.set(id, node);
        };
        /**
         * Lifecycles
         */
        ProjectionNode.prototype.mount = function (instance, isLayoutDirty) {
            var _a;
            if (isLayoutDirty === void 0) { isLayoutDirty = false; }
            if (this.instance)
                return;
            this.instance = instance;
            (_a = this.parent) === null || _a === void 0 ? void 0 : _a.children.add(this);
            this.root.potentialNodes.delete(this.id);
            if (isLayoutDirty) {
                this.isLayoutDirty = true;
                this.setTargetDelta(createDelta());
            }
            attachResizeListener === null || attachResizeListener === void 0 ? void 0 : attachResizeListener(instance, function () {
                // TODO: Complete all active animations/delete all projections
            });
        };
        // Note: currently only running on root node
        ProjectionNode.prototype.startUpdate = function () {
            this.isUpdating = true;
            // TODO: Traverse the tree, reset rotations
        };
        ProjectionNode.prototype.willUpdate = function (shouldNotifyListeners) {
            var _a;
            if (shouldNotifyListeners === void 0) { shouldNotifyListeners = true; }
            if (!this.isLead() || this.isLayoutDirty)
                return;
            this.isLayoutDirty = true;
            !this.root.isUpdating && this.root.startUpdate();
            this.path.forEach(function (node) {
                node.shouldResetTransform = true;
                /**
                 * TODO: Check we haven't updated the scroll
                 * since the last didUpdate
                 */
                node.updateScroll();
            });
            this.updateSnapshot();
            shouldNotifyListeners && ((_a = this.layoutWillUpdateListeners) === null || _a === void 0 ? void 0 : _a.notify());
        };
        // Note: Currently only running on root node
        ProjectionNode.prototype.didUpdate = function () {
            this.potentialNodes.forEach(function (node, id) {
                var element = document.querySelector("[data-projection-id=\"" + id + "\"]");
                if (element)
                    node.mount(element, true);
                console.log(!!element, id);
            });
            this.potentialNodes.clear();
            /**
             * Write
             */
            resetTreeTransform(this);
            /**
             * Read ==================
             */
            // Update layout measurements of updated children
            updateTreeLayout(this);
            /**
             * Write
             */
            // Notify listeners that the layout is updated
            notifyLayoutUpdate(this);
            // Flush any scheduled updates
            flushSync.update();
            flushSync.preRender();
            flushSync.render();
            this.isUpdating = false;
        };
        ProjectionNode.prototype.scheduleUpdateProjection = function () {
            sync.preRender(this.updateProjection, false, true);
        };
        /**
         * Update measurements
         */
        ProjectionNode.prototype.updateSnapshot = function () {
            if (this.snapshot)
                return;
            var visible = this.removeTransform(this.measure());
            this.snapshot = {
                visible: visible,
                layout: this.removeElementScroll(visible),
            };
        };
        ProjectionNode.prototype.updateLayout = function () {
            // TODO: Incorporate into a forwarded
            // scroll offset
            if (this.options.shouldMeasureScroll) {
                this.updateScroll();
            }
            if (!this.isLayoutDirty)
                return;
            this.layout = this.removeElementScroll(this.measure());
            this.layoutCorrected = createBox();
            this.isLayoutDirty = false;
        };
        ProjectionNode.prototype.updateScroll = function () {
            if (!measureScroll)
                return;
            this.scroll = measureScroll(this.instance);
        };
        ProjectionNode.prototype.resetTransform = function () {
            var _a, _b;
            if (resetTransform &&
                (this.isLayoutDirty || this.shouldResetTransform) &&
                ((this.projectionDelta && !isDeltaZero(this.projectionDelta)) ||
                    hasTransform(this.latestValues))) {
                resetTransform(this.instance);
                this.shouldResetTransform = false;
                (_b = (_a = this.options).onProjectionUpdate) === null || _b === void 0 ? void 0 : _b.call(_a);
            }
        };
        ProjectionNode.prototype.measure = function () {
            if (!measureViewportBox)
                return createBox();
            var box = measureViewportBox(this.instance);
            // Remove window scroll to give page-relative coordinates
            var scroll = this.root.scroll;
            // TODO Make loop
            scroll && eachAxis(function (axis) { return translateAxis(box[axis], scroll[axis]); });
            return box;
        };
        ProjectionNode.prototype.removeElementScroll = function (box) {
            var boxWithoutScroll = createBox();
            copyBoxInto(boxWithoutScroll, box);
            /**
             * Performance TODO: Keep a cumulative scroll offset down the tree
             * rather than loop back up the path.
             */
            for (var i = 0; i < this.path.length; i++) {
                var node = this.path[i];
                var scroll_1 = node.scroll, options = node.options;
                if (node !== this.root &&
                    scroll_1 &&
                    options.shouldMeasureScroll) {
                    translateAxis(boxWithoutScroll.x, scroll_1.x);
                    translateAxis(boxWithoutScroll.y, scroll_1.y);
                }
            }
            return boxWithoutScroll;
        };
        ProjectionNode.prototype.removeTransform = function (box) {
            var boxWithoutTransform = createBox();
            copyBoxInto(boxWithoutTransform, box);
            for (var i = 0; i < this.path.length; i++) {
                var node = this.path[i];
                if (!hasTransform(node.latestValues))
                    continue;
                hasScale(node.latestValues) && node.updateSnapshot();
                removeBoxTransforms(boxWithoutTransform, node.latestValues, node.snapshot.layout);
            }
            removeBoxTransforms(boxWithoutTransform, this.latestValues);
            return boxWithoutTransform;
        };
        /**
         *
         */
        ProjectionNode.prototype.setTargetDelta = function (delta) {
            this.targetDelta = delta;
            this.root.scheduleUpdateProjection();
        };
        ProjectionNode.prototype.setOptions = function (options) {
            this.options = options;
            var layoutId = this.options.layoutId;
            if (layoutId) {
                this.root.registerSharedNode(layoutId, this);
            }
        };
        /**
         * Frame calculations
         */
        ProjectionNode.prototype.resolveTargetDelta = function () {
            if (!this.targetDelta || !this.layout)
                return;
            if (!this.target) {
                this.target = createBox();
                this.targetWithTransforms = createBox();
            }
            copyBoxInto(this.target, this.layout);
            applyBoxDelta(this.target, this.targetDelta);
        };
        ProjectionNode.prototype.calcProjection = function () {
            var target = this.getLead().target;
            if (!this.layout || !target)
                return;
            if (!this.projectionDelta) {
                this.projectionDelta = createDelta();
                this.projectionDeltaWithTransform = createDelta();
            }
            /**
             * Reset the corrected box with the latest values from box, as we're then going
             * to perform mutative operations on it.
             */
            copyBoxInto(this.layoutCorrected, this.layout);
            /**
             * Apply all the parent deltas to this box to produce the corrected box. This
             * is the layout box, as it will appear on screen as a result of the transforms of its parents.
             */
            applyTreeDeltas(this.layoutCorrected, this.treeScale, this.path);
            /**
             * Update the delta between the corrected box and the target box before user-set transforms were applied.
             * This will allow us to calculate the corrected borderRadius and boxShadow to compensate
             * for our layout reprojection, but still allow them to be scaled correctly by the user.
             * It might be that to simplify this we may want to accept that user-set scale is also corrected
             * and we wouldn't have to keep and calc both deltas, OR we could support a user setting
             * to allow people to choose whether these styles are corrected based on just the
             * layout reprojection or the final bounding box.
             */
            calcBoxDelta(this.projectionDelta, this.layoutCorrected, target, this.latestValues);
            // TODO Make this event listener
            var onProjectionUpdate = this.options.onProjectionUpdate;
            onProjectionUpdate && onProjectionUpdate();
        };
        ProjectionNode.prototype.hide = function () {
            this.isVisible = false;
            // TODO: Schedule render
        };
        ProjectionNode.prototype.show = function () {
            this.isVisible = true;
            // TODO: Schedule render
        };
        ProjectionNode.prototype.scheduleRender = function () {
            var _a, _b;
            // TODO Rename this option
            (_b = (_a = this.options).onProjectionUpdate) === null || _b === void 0 ? void 0 : _b.call(_a);
        };
        ProjectionNode.prototype.getProjectionStyles = function () {
            // TODO: Return lifecycle-persistent object
            var styles = {};
            if (!this.isVisible) {
                return { visibility: "hidden" };
            }
            if (!this.projectionDelta) {
                return styles;
            }
            // Resolve crossfading props and viewport boxes
            // TODO: Return persistent mutable object
            this.applyTransformsToTarget();
            styles.transform = buildProjectionTransform(this.projectionDeltaWithTransform, this.treeScale, this.latestValues);
            if (this.instance.id === "box-a")
                console.log("rendering", styles.transform);
            // TODO Move into stand-alone, testable function
            var _a = this.projectionDelta, x = _a.x, y = _a.y;
            styles.transformOrigin = x.origin * 100 + "% " + y.origin * 100 + "% 0";
            /**
             * Apply scale correction
             */
            for (var key in scaleCorrectors) {
                if (this.latestValues[key] === undefined) {
                    delete styles[key];
                    continue;
                }
                var _b = scaleCorrectors[key], correct = _b.correct, applyTo = _b.applyTo;
                var corrected = correct(this.latestValues[key], this);
                if (applyTo) {
                    var num = applyTo.length;
                    for (var i = 0; i < num; i++) {
                        styles[applyTo[i]] = corrected;
                    }
                }
                else {
                    styles[key] = corrected;
                }
            }
            return styles;
        };
        ProjectionNode.prototype.applyTransformsToTarget = function () {
            var _a = this.getLead(), targetWithTransforms = _a.targetWithTransforms, target = _a.target, latestValues = _a.latestValues;
            copyBoxInto(targetWithTransforms, target);
            /**
             * Apply the latest user-set transforms to the targetBox to produce the targetBoxFinal.
             * This is the final box that we will then project into by calculating a transform delta and
             * applying it to the corrected box.
             */
            transformBox(targetWithTransforms, latestValues);
            /**
             * Update the delta between the corrected box and the final target box, after
             * user-set transforms are applied to it. This will be used by the renderer to
             * create a transform style that will reproject the element from its actual layout
             * into the desired bounding box.
             */
            calcBoxDelta(this.projectionDeltaWithTransform, this.layoutCorrected, targetWithTransforms, latestValues);
        };
        /**
         * Events
         *
         * TODO Replace this with a key-based lookup
         */
        ProjectionNode.prototype.onLayoutWillUpdate = function (handler) {
            if (!this.layoutWillUpdateListeners) {
                this.layoutWillUpdateListeners = new SubscriptionManager();
            }
            return this.layoutWillUpdateListeners.add(handler);
        };
        ProjectionNode.prototype.onLayoutDidUpdate = function (handler) {
            if (!this.layoutDidUpdateListeners) {
                this.layoutDidUpdateListeners = new SubscriptionManager();
            }
            return this.layoutDidUpdateListeners.add(handler);
        };
        ProjectionNode.prototype.registerSharedNode = function (layoutId, node) {
            if (!this.sharedNodes.has(layoutId)) {
                this.sharedNodes.set(layoutId, new NodeStack());
            }
            var stack = this.sharedNodes.get(layoutId);
            stack.add(node);
        };
        ProjectionNode.prototype.isLead = function () {
            var stack = this.getStack();
            return stack ? stack.lead === this : true;
        };
        ProjectionNode.prototype.getLead = function () {
            var _a;
            var layoutId = this.options.layoutId;
            return layoutId ? ((_a = this.getStack()) === null || _a === void 0 ? void 0 : _a.lead) || this : this;
        };
        ProjectionNode.prototype.getStack = function () {
            var layoutId = this.options.layoutId;
            if (layoutId)
                return this.root.sharedNodes.get(layoutId);
        };
        ProjectionNode.prototype.promote = function (options) {
            var stack = this.getStack();
            if (stack)
                stack.promote(this, options);
        };
        return ProjectionNode;
    }());
}
function updateTreeLayout(node) {
    node.updateLayout();
    node.children.forEach(updateTreeLayout);
}
function notifyLayoutUpdate(node) {
    var layout = node.layout, snapshot = node.snapshot;
    if (layout && snapshot && node.layoutDidUpdateListeners) {
        // TODO Maybe we want to also resize the layout snapshot so we don't trigger
        // animations for instance if layout="size" and an element has only changed position
        if (node.options.animationType === "size") {
            eachAxis(function (axis) {
                var axisSnapshot = snapshot.visible[axis];
                var length = calcLength(axisSnapshot);
                axisSnapshot.min = layout[axis].min;
                axisSnapshot.max = axisSnapshot.min + length;
            });
        }
        else if (node.options.animationType === "position") {
            eachAxis(function (axis) {
                var axisSnapshot = snapshot.visible[axis];
                var length = calcLength(layout[axis]);
                axisSnapshot.max = axisSnapshot.min + length;
            });
        }
        var layoutDelta = createDelta();
        calcBoxDelta(layoutDelta, layout, snapshot.layout);
        var visualDelta = createDelta();
        calcBoxDelta(visualDelta, layout, snapshot.visible);
        node.layoutDidUpdateListeners.notify({
            layout: layout,
            snapshot: snapshot,
            delta: visualDelta,
            hasLayoutChanged: !isDeltaZero(layoutDelta),
        });
    }
    node.children.forEach(notifyLayoutUpdate);
    node.snapshot = undefined;
}
function resetTreeTransform(node) {
    node.resetTransform();
    node.children.forEach(resetTreeTransform);
}
function updateProjectionTree(node) {
    resolveTreeTargetDeltas(node);
    calcTreeProjection(node);
}
function resolveTreeTargetDeltas(node) {
    node.resolveTargetDelta();
    node.children.forEach(resolveTreeTargetDeltas);
}
function calcTreeProjection(node) {
    node.calcProjection();
    node.children.forEach(calcTreeProjection);
}

export { createProjectionNode };
